from TDhelper.network.http.REST_HTTP import POST, GET
from datetime import datetime, timedelta, timezone
import json


class token_manage:
    __token = None
    __expire = None
    __token_service_uri = None
    __service_key = None
    __service_secret = None
    __time_cls = None

    @property
    def Token(self):
        if not self.__token:
            self.__get_token__()
        if self.__expire:
            if datetime.strptime(self.__expire,"%Y-%m-%d %H:%M:%S.%f") < self.__time_cls.now():
                self.__get_token__()
        return self.__token

    def __init__(self, token_uri:str, svr_key:str, secret:str, time_cls=None) -> None:
        self.__time_cls = datetime if not time_cls else time_cls
        self.__service_key = svr_key
        self.__service_secret = secret
        token_uri= token_uri.rstrip('/')
        self.__token_service_uri = "/".join([token_uri, "api/rpc_token"])
        self.__get_token__()

    def __get_token__(self):
        try:
            state, body = GET(self.__token_service_uri, {
                "svr_key": self.__service_key.upper(),
                "secret": self.__service_secret
            })
            if state:
                data = json.loads(body)
                self.__token = data.get("msg").get("token", None)
                self.__expire = data.get("msg").get("expire", None).replace("T"," ")
        except Exception as e:
            raise e



#! python 3.6

#########################################################
# | @File    :   Service.py
# | @ClsName :   Service
# | @Version :   1.0.0
# | @History :

# you can input 'p-history-table' to insert a history record.
# |----------------------------------------------------------------------------------------------------------|
# | version       | Type                | Author        | Contact                     | Time                 |
# |----------------------------------------------------------------------------------------------------------|
# | 1.0.0         | Create              | Tony.Don      | yeihizhi@163.com            | 2022-09-29 09:27:43  |
# |----------------------------------------------------------------------------------------------------------|
# | @License :   Apache 2.0
# | @Desc    :   RPC server service.
#########################################################

# import lib by pypi

from copy import deepcopy
import json
import logging
from TDhelper.generic.dictHelper import findInDict, createDictbyStr
from TDhelper.network.rpc.Generic.Host import HostManager
from TDhelper.network.http.REST_HTTP import GET, POST, PUT, DELETE, ContentType
# import lib by project
# code start


class override_Attribute:

    def __init__(self, name, map=None):
        self.__name__ = name
        if map:
            self.__name__ = map+"."+self.__name__

    def __set__(self, instance, v):
        createDictbyStr(self.__name__, instance.__dict__, v)

    def __get__(self, instance, owen):
        try:
            return findInDict(self.__name__, instance.__dict__)
        except Exception as e:
            return None

    def __delete__(self, instance):
        name = self.__name__.split('.')
        name = name[len(name)-1]
        instance.__dict__[self.__map__].pop(name)


class Service:
    __cache__ = None
    __conf__ = None
    __logger_hander__ = None
    __ENCODING__ = "UTF-8"
    __HEADERS__ = {}
    __HOST__ = HostManager()
    __USR_ENCODING__ = override_Attribute('encoding', "__conf__")
    __HOST_CONF__ = override_Attribute('hosts', "__conf__")
    __TOKEN_KEY__ = override_Attribute('token_key', "__conf__")
    __SECRET__ = override_Attribute('secret', "__conf__")
    __SERVICE_PATH__ = override_Attribute('uris.service', '__conf__')
    __METHOD_PATH__ = override_Attribute('uris.method', '__conf__')
    __HOST_PATH__ = override_Attribute('uris.host', '__conf__')
    __PARAMS_PATH__ = override_Attribute("uris.param", '__conf__')
    __RETURN_PATH__ = override_Attribute('uris.return', '__conf__')
    __RETURN_DESC_PATH__ = override_Attribute("uris.return_desc", '__conf__')

    def __init__(self, conf, cache_handle={}, logger=None) -> None:
        self.__cache__ = cache_handle
        self.__conf__ = conf
        self.__HEADERS__ = {
            self.__TOKEN_KEY__: self.__SECRET__
        }
        if self.__USR_ENCODING__:
            self.__ENCODING__ = self.__USR_ENCODING__
        self.__logger_hander__ = logger if logger else logging
        offset = 0
        for item in self.__HOST_CONF__:
            self.__HOST__.register(
                serverId=offset, host=item['host'], port=80 if 'port' not in item else item['port'], proto=item['protocol'], status=True)
            offset += 1

    def getService(self, key):
        return self.__get_service__(key)

    def __get_service__(self, key=None):
        if key not in self.__cache__:
            post_data={"key":key} if key else b''
            self.__HOST__.getHost()
            statu, body = self.__remote__(self.__get_uri__(self.__SERVICE_PATH__),post_data,deepcopy(self.__HEADERS__))
            if statu:
                body= json.loads(body)
                if body["state"]==200:
                    for k,v in body["msg"].items():
                        body["msg"][k]["hosts"]=self.__get_host__(v["id"])
                        body["msg"][k]["methods"]=self.__get_method__(v["id"])
                        self.__cache__[k]=v
        if key:
            if key in self.__cache__:
                return self.__cache__[key]
        else:
            return self.__cache__

    def __get_host__(self,s_id):
        get_data={"s_id":s_id}
        state,body=self.__remote__(self.__get_uri__(self.__HOST_PATH__),get_data,deepcopy(self.__HEADERS__))
        if state:
            body= json.loads(body)
            if body["state"]==200:
                return body['msg']
            else:
                return []
        else:
            return []

    def __get_method__(self,s_id,key=None):
        get_data= {"s_id":s_id}
        if key:
            get_data['key']=key
        state,body=self.__remote__(self.__get_uri__(self.__METHOD_PATH__),get_data,deepcopy(self.__HEADERS__))
        if state:
            body= json.loads(body)
            if body["state"]==200:
                for item in body['msg']:
                    for k,v in item.items():
                        item[k]['params']=self.__get_method_params__(v['id'])
                        item[k]['returns']=self.__get_return__(v['id'])
                return body['msg']
            else:
                return []
        else:
            return []

    def __get_method_params__(self,m_id):
        get_data={"m_id":m_id}
        state,body=self.__remote__(self.__get_uri__(self.__PARAMS_PATH__),get_data,deepcopy(self.__HEADERS__))
        if state:
            body=json.loads(body)
            if body["state"]==200:
                return body["msg"]
            else:
                return []
        else:
            return []

    def __get_return__(self,m_id):
        get_data={"m_id":m_id}
        state,body= self.__remote__(self.__get_uri__(self.__RETURN_PATH__),get_data,deepcopy(self.__HEADERS__))
        if state:
            body= json.loads(body)
            if body["state"]==200:
                body['msg']['descriptions']= self.__get_return_desc__(body['msg']['id'])
                return body['msg']
            else:
                return {}
        else:
            return {}
    
    def __get_return_desc__(self,r_id):
        get_data={"r_id":r_id}
        state,body=self.__remote__(self.__get_uri__(self.__RETURN_DESC_PATH__),get_data,deepcopy(self.__HEADERS__))
        if state:
            body= json.loads(body)
            if body["state"]==200:
                return body['msg']
            else:
                return []
        else:
            return []

    def __get_uri__(self, uri):
        uri = uri.strip('/')
        return ''.join([self.__HOST__.__generateHost__(self.__HOST__.getHost()), "/", uri, "/"])

    def __remote__(self, uri, data=b"", headers={}):
        try:
            state, res = GET(uri, post_data=data,
                             http_headers=headers, content_type=ContentType.JSON, time_out=5)
            if state == 200:
                return True, str(res, self.__ENCODING__)
            else:
                self.__logger_hander__.error(
                    "Access('%s'), HTTP CODE: %s" % (uri, state))
                return False, state
        except Exception as e:
            self.__logger_hander__.error(e.args)
            return False, state


if __name__ == "__main__":
    cc = Service({
        "encoding": "utf-8",
        "token_key": "api-token",
        "secret": "d74218cff240f446d341c4a2f3a8c588",
        "hosts": [
            {
                "host": "192.168.50.2",
                "port": "10001",
                "protocol": "http://"
            }
        ],
        "uris": {
            "service": "api/rpcb/",
            "method": "api/rpc_method/",
            "host": "api/rpc_host/",
            "param": "api/rpc_method_params/",
            "return": "api/rpc_return/",
            "return_desc": "api/rpc_return_desc/"
        }
    })
    print(cc.__get_service__())
    print("gg")


import functools
import logging
import logging.config

logging.basicConfig(level= logging.INFO)

def logging_setup(config_json):
    logging.config.dictConfig(config= config_json)

def SYS_LOGS(func):
    def wapper(*args, **kwargs):
        try:
            logging.info("{} start..." % func.__name__)
            ret= func(*args, **kwargs)
            logging.info("{} end..." % func.__name__)
            return ret
        except Exception:
            logging.exception("System Faild.",exc_info= True)
    return wapper

def log(func):
    def wapper(*args, **kwargs):
        try:
            logging.info('{} is start.' % func.__name__)
            ret= func(*args, **kwargs)
            logging.info('{} is end.' % func.__name__)
            return ret
        except Exception as e:
            logging.error("{} has error:" % func.__name__)
            logging.error("{}" % e)
    return wapper


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import bson
import copy
from bson.codec_options import CodecOptions
from TDhelper.db.mongodb.dbhelper import dbhelper, db_cfg

from TDhelper.Spider.cfg import setting

'''
class\r\n
    objectId
description\r\n
    mongodb's bson.objectid\r\n
'''


class objectId(dbhelper):
    model = None

    def __init__(self):
        super(objectId, self).__init__()
        self.setCollection(type(self).__name__)
        self.model = {
            "_id": None
        }

    @property
    def oId(self):
        if self.model["_id"]:
            return self.model["_id"]
        else:
            self.model["_id"] = bson.objectid.ObjectId()
            return self.model["_id"]

    @oId.setter
    def oId(self, args):
        if args:
            self.model["_id"] = bson.objectid.ObjectId(args)
        else:
            self.model["_id"] = bson.objectid.ObjectId()
    
    def clear(self):
        for item in self.model.keys():
            self.model[item]= None

    def toSave(self):
        if self.model:
            if self.oId:
                return self.save(self.model)
        else:
            return None

    def deleteById(self):
        if self.model:
            return self.remove({'_id': self.oId})

    def getbyId(self):
        if self.model:
            result = copy.deepcopy(self)
            if result:
                oResult = self.findOne({'_id': self.oId})
                if oResult:
                    result.model = oResult
                    return result
                return None
        return None

    def getByfield(self, field_name):
        if field_name:
            return self.findOne({field_name: self.model[field_name]})
        return None

    def UpdateById(self):
        if self.oId:
            self.update({'_id': self.oId}, self.model)

    def getOneByQuery(self, query):
        if self.model:
            result = copy.deepcopy(self)
            if result:
                oResult = self.findOne(query)
                if oResult:
                    result.model = oResult
                    return result
        return None



from types import FunctionType
from copy import copy
import os
import logging
from TDhelper.generic.requier import R as require
from TDhelper.generic.classDocCfg import doc
from TDhelper.network.rpc.Core.Meta import Meta
from TDhelper.network.rpc.Core.struct import *
from TDhelper.network.rpc.Core.token import token_manage
from TDhelper.network.http.REST_HTTP import GET, POST, PUT, DELETE, ContentType

# 反射获取类
def reflectClass(cls):
    return require(cls).getType()

class SRC(metaclass=Meta):
    """SRC
        service register center interface.
    Parameters:
        uris - <class:list>, rpc server apis.
        server_cnf - <class:dict>, service conf.
        rpc_conf_header - <class:dict>, rpc configure general header struct. example :{
                    "name": "权限服务",
                    "description": "权限服务",
                    "key": "RPC_TEST",
                    "secret": "203920398409238408204324",
                    "protocol": "http://",
                    "hosts": [
                        {
                            "serverId": 1,
                            "host": "192.168.0.1",
                            "port": 8080,
                            "sniffer": "api/sniffer",
                            "proto": "http://"
                        }
                    ]
                } 
        logger - <class:logging>, log handle.
        **kwargs:
            token_key - <class:str>, access token key. it will auto assembly http header. default 'api-token'.
    Returns:
        None
    Raises:
        None
    """
    def __init__(self, uris: list, server_cnf: dict = None, rpc_conf_header:dict={}, logger: "logging" = None,**kwargs):
        self.__RPC_CNF__ = RPC_SERVICE_CONF(uris)
        self.__token_manage__= token_manage(self.__RPC_CNF__.getUri(),rpc_conf_header.get("key"),rpc_conf_header.get('secret'))
        self.__logger_hander__ = logger if logger else self.__logger_hander__
        self.__rpc_server_cnf__ = server_cnf if server_cnf else self.__rpc_server_cnf__
        apiToken="api-token" if "token_key" not in kwargs else kwargs["token_key"]
        #self.__headers__[apiToken] = self.__token_manage__.Token#self.__RPC_CNF__.token
        self.__rpc_conf_header__= rpc_conf_header
        self.__headers__.update({apiToken:self.__token_manage__.Token})
        self.__headers__.update({"access-source":"rpc"})

    def __set_coding__(self, encoding='utf-8'):
        self.__ENCODING__ = encoding

    def __setting__(self, cnf):
        self.__rpc_server_cnf__ = cnf

    def AutoRegister(self,cxt:list,conf_header:dict= None):
        """AutoRegister
            Auto register service by method annotate.
        Parameters:
            cxt - <class:list>, want to auto register cls.
            conf_header - <class:dict>, override rpc configure general header struct.
        Returns:
            None
        Raises:
            None
        Example annotate:
            Conf In Annotate:
                [rpc]
                {
                    "uri": <class, str>,
                    "method": <class, enum(GET|POST|PUT|DELETE)>,
                    "version": <class, str(1.0.0)>,
                    "description": <class: str>,
                    "params": [
                        {
                            "key":<class, str>,
                            "description":<class, str>,
                            "defaultValue":<class, str>
                        }
                    ],
                    "returns":{
                        "valueType":"json",
                        "examples":<class, str>,
                        "descriptions":[
                            {
                                "key":<class, str>,
                                "valueDescription":<class, str>
                            }
                        ]
                    }
                }
                [rpcend]
            Conf In Json:
                [rpc]
                methods.fun1::test.json
                [rpcend]

                [rpc]
                fun1::test.json
                [rpcend]
        """
        conf= conf_header if conf_header else copy(self.__rpc_conf_header__)
        if not conf:
            raise Exception("configure header is none. see more in config/regist_conf/readme.md")
        if "methods" not in conf:
            conf['methods']={}
        for cxt_item in cxt:
            m_cxt= cxt_item
            if isinstance(cxt_item,str):
                m_cxt= reflectClass(cxt_item)
            for k, v in m_cxt.__dict__.items():
                if isinstance(v, FunctionType):
                    func_name = "".join([m_cxt.__name__,'_',v.__name__]).upper()
                    if v.__doc__:
                        methods = doc(v.__doc__, "rpc")
                        if methods:
                            methods = methods.replace("\n", "").strip()
                            try:
                                if func_name not in conf['methods']:
                                    conf['methods'][func_name]=json.loads(methods)#, encoding="utf-8")
                                    conf['methods'][func_name]["collect"]=m_cxt.__name__
                            except Exception as e:
                                raise e
        self.__register_by_cnf__(conf)
        return self
    
    def Register(self, cxt:"dict|str", filepath:bool=False, conf_header:dict=None):
        """Register
            register service by conf.
        Parameters:
            cxt - <class:dict|str>, conf json or conf file path.
            filepath - <class:bool>, control conf mode. default False use dict to register.
            conf_header - <class:dict>, override rpc configure general header struct.
        Returns:
            None
        Raises:
            None
        """
        conf= conf_header if conf_header else copy(self.__rpc_conf_header__)
        if not conf:
            raise Exception("configure header is none. see more in config/regist_conf/readme.md")
        
        if filepath:
            if os.path.exists(cxt):
                m_conf= self.__register_by_file__(cxt)
                if "methods" in m_conf:
                    conf.update(m_conf)
                else:
                    conf.update({"methods":m_conf})
                self.__register_by_cnf__(conf)
            else:
                raise Exception("can not found path '%s'" % cxt)
        else:
            if "methods" in cxt:
                if "methods" in cxt:
                    conf.update(cxt)
                else:
                    conf.update({"methods":cxt})
            self.__register_by_cnf__(conf)
        return self

    def __register_by_file__(self, filepath):
        m_json_source = None
        filepath = filepath.replace("\\", "/")
        with open(filepath, mode='r', encoding=self.__ENCODING__) as f:
            m_json_source = f.read()
            f.close()
        return json.loads(m_json_source)

    def __register_by_cnf__(self, cnf: "dict|str"):
        if isinstance(cnf, str):
            cnf = json.loads(cnf)
        self.__analysis_cnf__(cnf)

    def __register_service__(self):
        uri = "".join([self.__RPC_CNF__.getUri(),
                      self.__service_register_uri__])
        state, res = self.__remote__(
            uri, self.__RPC_SERVICES__.toCnf(), copy(self.__headers__))
        if state:
            res = json.loads(res)
            if res["state"] == 200:
                self.__logger_hander__.info("service(%s) register success."%res['msg']['name'])
                s_id = res["msg"]["id"]
                self.__register_host__(s_id)
                self.__register_method__(s_id)
            else:
                self.__logger_hander__.error(
                    "Access(%s), '%s'" % (uri, res["msg"]))

    def __register_host__(self, parentId):
        uri = "".join([self.__RPC_CNF__.getUri(), self.__host_register_uri__])
        for v in self.__RPC_SERVICES__.host.__host__:
            host_cnf = self.__RPC_SERVICES__.host.get_host_by_key(v)
            data = {
                "host": host_cnf['host'],
                "port": host_cnf['port'],
                "state": True,
                "service": parentId,
                "sniffer":host_cnf['sniffer']
            }
            state, res = self.__remote__(uri, data, copy(self.__headers__))
            if state:
                res = json.loads(res)
                if res["state"] == 200:
                    self.__logger_hander__.info(
                        "register host '%s' success." % host_cnf['host'])
                else:
                    self.__logger_hander__.error(
                        "register host error(%s)" % res["msg"])

    def __register_method__(self, parentId):
        uri = "".join([self.__RPC_CNF__.getUri(),
                      self.__method_register_uri__])
        for k,v in self.__RPC_SERVICES__.methods.items():
            data = v
            data.service = parentId
            state, res = self.__remote__(
                uri, data.toCnf(), copy(self.__headers__))
            if state:
                res = json.loads(res)
                if res["state"] == 200:
                    self.__logger_hander__.info(
                        "register method '%s' success." % v.name)
                    m_id = res["msg"]["id"]
                    self.__register_parameters__(v, m_id)
                    self.__register_returns__(v, m_id)
                else:
                    self.__logger_hander__.error(
                        "register method '%s' error (%s)" % (v.name, res["msg"]))

    def __register_parameters__(self, method, parentId):
        uri = "".join([self.__RPC_CNF__.getUri(),
                      self.__params_register_uri__])
        for item in method.params:
            data = item
            data.serviceUri = parentId
            state, res = self.__remote__(
                uri, data.toCnf(), copy(self.__headers__))
            if state:
                res = json.loads(res)
                if res["state"] == 200:
                    self.__logger_hander__.info(
                        "register method paramete '%s' success." % item.key)
                else:
                    self.__logger_hander__.error(
                        "register method %s error(%s)." % (item.key, res["msg"]))

    def __register_returns__(self, method, parentId):
        uri = "".join([self.__RPC_CNF__.getUri(),
                      self.__return_register_uri__])
        data = method.returns
        data.serviceUri = parentId
        state, res = self.__remote__(uri, data.toCnf(), copy(self.__headers__))
        if state:
            res = json.loads(res)
            if res["state"] == 200:
                self.__logger_hander__.info("register method return success.")
                r_id = res["msg"]["id"]
                self.__register_return_desc__(data.descriptions, r_id)
            else:
                self.__logger_hander__.error(
                    "register method return error(%s)" % res["msg"])

    def __register_return_desc__(self, items, return_id):
        uri = "".join(
            [self.__RPC_CNF__.getUri(), self.__return_desc_register_uri__])
        for item in items:
            data = item
            data.returns = return_id
            state, res = self.__remote__(
                uri, data.toCnf(), copy(self.__headers__))
            if state:
                res = json.loads(res)
                if res["state"] == 200:
                    self.__logger_hander__.info(
                        "register method return description '%s' success." % item.key)
                else:
                    self.__logger_hander__.error(
                        "register method return description '%s' error(%s)", (item.key, res["msg"]))

    def __analysis_cnf__(self, cnf):
        assert isinstance(cnf, dict), "cnf must dict types."
        try:
            self.__RPC_SERVICES__ = RPC_SERVICE.create_by_cnf(cnf)
            if self.__RPC_SERVICES__:
                self.__register_service__()
            else:
                raise Exception("RPC services create error.")
        except Exception as e:
            raise e

    def __sniffer__(self):
        return True, self.__rpc_server_cnf__["key"]

    def __remote__(self, uri, data=b"", headers={}):
        try:
            state, res = POST(uri, post_data=data,
                              http_headers=headers, content_type=ContentType.JSON, time_out=5)
            if state == 200:
                return True, str(res, self.__ENCODING__)
            else:
                self.__logger_hander__.error(
                    "Access('%s'), HTTP CODE: %s" % (uri, state))
                return False, None
        except Exception as e:
            self.__logger_hander__.error(e.args)
            return False, None


import mysqlx

class mysql_x:
    

    def __init__(self, connection_str, options_str):
        self.__conn_str= None
        self.__options_str= None
        self.clients= None

    def getSession(self):
        if self.clients:
            return self.clients.get_session()

    def createSessionPool(self, connection_str, options_str):
        self.clients= mysqlx.get_client(connection_str, options_str)

    def openDataBase(self, dataBaseName, session):
        if session:
            return session.get_schema(dataBaseName)

    def openTable(self, tableName, schema):
        if schema:
            return schema.get_collection(tableName)

def getConnectionStr(host,port,user,password):
    return {
            'host':host if host else "host.5ker.com",
            'port':port if port else 3308,
            'user':user,
            'password':password
    }

def getPoolOptionsStr(enabled=True,maxSize=1,max_idle_time=0,queue_timeout=0):
    return {
            'pooling':{
            'enabled':enabled,
            'max_size':maxSize,
            'max_idle_time':max_idle_time,
            'queue_timeout':queue_timeout
        }
    }
